/*
 * xtract V 1.0 - xdvdfs extraction utility
 *
 * This tool is based on xbiso v0.56
 * Copyright (C) 2002  Tonto Rostenfaunt   xbiso@linuxmail.org
 *
 */

#define _FILE_OFFSET_BITS	64

#include <features.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include "xtract.h"

int	listonly,
	debug;

void dumpde(struct dirent *de) {
	char	filename[256];
	printf("Dump of xiso directory entry:\n");
	printf("\tltable  : %04x\n", de->ltable);
	printf("\trtable  : %04x\n", de->rtable);
	printf("\tsector  : %08x\n", de->sector);
	printf("\tsize    : %08x\n", de->size);
	printf("\tattrib  : %02x\n", de->attribs);
	printf("\tnamelen : %02x\n", de->namelen);

	strncpy((char *) &filename,
			(const char *) &de->filename,
			MIN(de->namelen, 255));
	filename[MIN(de->namelen, 255)]=0x0;
	
	printf("\tfilename: %s\n", filename);
}

void help(int ecode) {

	printf("\nxxtract Copyright 2003 Florian Lohoff <flo@rfc822.org> \n\n");
	printf("\t-d dirname\tdirectory to extract to\n");
	printf("\t-l\t\tlist only\n");
	printf("\t-D\t\tdebug\n");

	exit(ecode);
}

void err(char *error) {
	printf("%s", error);
	exit(1);
}

void xextract(int xfd, long sector, long size, char *dir) {

	struct dtable	dt[XISO_DIR_MAXDEPTH];
	char		filename[256];
	char		absfile[2048];
	char		*cbuf;
	int		depth=0,
			offset,
			debroken;

	dt[depth].sector=sector;
	dt[depth].size=size;
	dt[depth].buf=malloc(size);
	dt[depth].dir=strdup(dir);

	if (dt[depth].buf == NULL || dt[depth].dir == NULL) 
		err("Malloc failed\n");

	if ((lseek(xfd, sector*XISO_SECTOR_SIZE, SEEK_SET)) == -1)
		err("Seek failed to root directory - xiso damaged ?\n");

	if (read(xfd, dt[depth].buf, size) != size) 
		err("Read of root directory failed - xiso damaged ?\n");

	dt[depth].de=(struct dirent *) dt[depth].buf;

	if (!(cbuf=malloc(COPY_BUFFER_SIZE))) 
		err("Malloc failed\n");

	while(1) {
		debroken=0;

		/* Check for brokeness of the direntrys which
		 * should not be processed further */
		if (dt[depth].de->ltable == 0xffff ||
			dt[depth].de->rtable == 0xffff ||
			dt[depth].de->size == 0xffffffff ||
			dt[depth].de->sector == 0xffffffff ||
			dt[depth].de->attribs == 0xff) 
			debroken=1;

		if (debug)
			dumpde(dt[depth].de);
	
		strncpy((char *) &filename,
			(const char *) &dt[depth].de->filename,
			dt[depth].de->namelen);

		filename[(int) dt[depth].de->namelen]=0x0;

		sprintf(absfile, "%s/%s", dt[depth].dir, filename);

		if (debroken) {
			dumpde(dt[depth].de);
			printf("Broken direntry in xiso .. skipping dir\n");
		} else if (dt[depth].de->attribs & XISO_ATTRIB_DIR) {
			printf("0x%02x              %s/\n",
					dt[depth].de->attribs,
					absfile);

			if (!listonly)
				mkdir(absfile, 0755);

			/* Process only directories with files */
			if (dt[depth].de->sector != 0 &&
				dt[depth].de->size != 0) {

				dt[depth+1].sector=dt[depth].de->sector;
				dt[depth+1].size=dt[depth].de->size;
				dt[depth+1].buf=malloc(dt[depth].de->size);
				dt[depth+1].dir=strdup(absfile);

				if (dt[depth+1].buf == 0 || dt[depth+1].dir == 0)
					err("Malloc failed\n");

				depth++;

				if (lseek(xfd, dt[depth].sector*XISO_SECTOR_SIZE, SEEK_SET) == -1) {
					dumpde(dt[depth-1].de);
					err("Seek to directory failed - xiso damaged ?\n");
				}

				if (read(xfd, dt[depth].buf, dt[depth].size) != dt[depth].size) {
					dumpde(dt[depth-1].de);
					err("Short read on reading directory structure - xiso damaged ?\n");
				}

				dt[depth].de=(struct dirent *) dt[depth].buf;

				continue;
			}
		} else if (dt[depth].de->attribs & (XISO_ATTRIB_FILE|XISO_ATTRIB_ARC)) {
			int	of,left;

			printf("0x%02x %12d %-s\n",
				dt[depth].de->attribs, 
				dt[depth].de->size,
				absfile);

			if (!listonly) {
				if ((of=open(absfile, O_CREAT|O_RDWR, 0644)) == -1)
					err("Failed opening output file\n");

				if (lseek(xfd, dt[depth].de->sector*XISO_SECTOR_SIZE, SEEK_SET) == -1)
					err("Seek to file failed - xiso damaged ?\n");

				for(left=dt[depth].de->size;left>0;) {
					int	copynow=MIN(COPY_BUFFER_SIZE, left);

					if (read(xfd, cbuf, copynow) != copynow)
						err("Short read in xiso file\n");
					if (write(of, cbuf, copynow) != copynow)
						err("Short write in output file\n");

					left-=copynow;
				}

				close(of);
			}
		} else {
			printf("Unknown file attribs %02x\n", dt[depth].de->attribs);
			dumpde(dt[depth].de);
		}

		/* Check for brokeness */
		if ((dt[depth].de->rtable*4) >= (dt[depth].size-XISO_MIN_DESIZE)) 
			debroken=1;

		/* While not root directory end */
		while((dt[depth].de->ltable == 0 && dt[depth].de->rtable == 0) || debroken) {

			free(dt[depth].buf);
			free(dt[depth].dir);

			/* We are at the end of the root directory ? */
			if (!depth)
				return;

			depth--;
			debroken=0;
		}

		offset=dt[depth].de->rtable;
		dt[depth].de=(struct dirent *) (((long *) dt[depth].buf) + offset);
	}
}

int main(int argc, char *argv[]) {

	struct	vdescr	*vdbuf;
	char	*dbuf=NULL,
		*fname=NULL;

	int	ret,
		xfd=0;

	while((ret = getopt(argc,argv,"Dhld:")) != -1) {
		switch(ret) {
			case 'd':
				dbuf=strdup(optarg);
				break;
			case 'l':
				listonly=1;
				break;
			case 'D':
				debug=1;
				break;
			case 'h':
				help(0);
 				break;
		}
	}

	if (optind >= argc) 
		help(1);

	fname=strdup(argv[optind]);

	/* set the dirname to the filename - ext if blank */
	/* FIXME: Check for .iso at end - otherwise don't delete */
	if(dbuf==NULL) {
		int 	i,	
			l=strlen(fname);

		/* Duplicate string */
		dbuf=strdup(fname);

		/* Delete last 4 chars */
		for(i=l;l-4<=i;i--) 
			dbuf[i]=0x0;
	}

	/* FIXME: Should check for 0 */
	vdbuf=(struct vdescr *) malloc(sizeof(struct vdescr));

	if ((xfd=open(fname, O_RDONLY)) == -1)
		err("Could not open xiso file\n");

	if(!listonly)
		if(mkdir(dbuf, 0755) != 0)
			err("Failed to create root directory\n");

	lseek(xfd, XISO_VD_SECT*XISO_SECTOR_SIZE, SEEK_SET);
	read(xfd, vdbuf, sizeof(struct vdescr));

	if((strncmp(vdbuf->idstring, XISO_VDESCR_IDSTRING,
				strlen(XISO_VDESCR_IDSTRING))) != 0)
		err("Error file doesnt appear to be a xbox iso image\n");

	xextract(xfd, vdbuf->rootsector, vdbuf->rootsize, dbuf);

	return 0;
}
